from __future__ import annotations

__version__ = '0.2.1'

# transformed numFiles to string, sizeInBytes -> size with type string
from datetime import datetime
from time import time
import pytz

DETAIL_SCHEMA_JSON = '{"fields":[{"metadata":{},"name":"createdAt","nullable":true,"type":"timestamp"},{"metadata":{},"name":"description","nullable":true,"type":"string"},{"metadata":{},"name":"format","nullable":true,"type":"string"},{"metadata":{},"name":"id","nullable":true,"type":"string"},{"metadata":{},"name":"lastModified","nullable":true,"type":"timestamp"},{"metadata":{},"name":"location","nullable":true,"type":"string"},{"metadata":{},"name":"minReaderVersion","nullable":true,"type":"long"},{"metadata":{},"name":"minWriterVersion","nullable":true,"type":"long"},{"metadata":{},"name":"name","nullable":true,"type":"string"},{"metadata":{},"name":"numFiles","nullable":true,"type":"string"},{"metadata":{},"name":"partitionColumns","nullable":true,"type":{"containsNull":true,"elementType":"string","type":"array"}},{"metadata":{},"name":"properties","nullable":true,"type":{"keyType":"string","type":"map","valueContainsNull":true,"valueType":"string"}},{"metadata":{},"name":"size","nullable":true,"type":"string"}],"type":"struct"}'  # noqa: E501


def _humanize_number(number: int) -> str:
    return f'{number:,}'


def _humanize_bytes(num_bytes: float) -> str:
    # ChatGPT 🤖 prompt: "write a python program that converts bytes to their proper units (kb, mb, gb, tb, pb, etc)" # noqa: E501
    units = ['bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB']
    i = 0
    while num_bytes >= 1024 and i < len(units) - 1:
        num_bytes /= 1024
        i += 1
    return f'{num_bytes:.2f} {units[i]}'


def _humanize_timestamp(ts: int, timezone: str = 'UTC') -> str:
    if ts > time() * 10:  # if the timestamp is greater than our current one by a factor of 10, the ts is in ms
        ts = ts / 1000
    tz = pytz.timezone(timezone)
    return datetime.fromtimestamp(ts, tz).isoformat()



from __future__ import annotations

import copy
import math

from delta import DeltaTable
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession

import hydro.spark as hs
from hydro import _humanize_bytes
from hydro import _humanize_number


def _summarize_data_files(delta_table: DeltaTable):
    base_path = delta_table.detail().collect()[0]['location']
    spark = delta_table.toDF().sparkSession
    hadoop_conf = spark._jsparkSession.sessionState().newHadoopConf()
    glob_path = spark._jvm.org.apache.hadoop.fs.Path(base_path)
    driver_fs = glob_path.getFileSystem(hadoop_conf)
    files = driver_fs.listFiles(glob_path, True)
    file_count = 0
    total_size = 0
    min_ts = math.inf
    while files.hasNext():
        file = files.next()
        path = file.getPath()
        if '_delta_log' in path.toUri().toString():  # this ain't great - figure out globbing instead of doing this.
            continue
        file_count += 1
        total_size += file.getLen()
        min_ts = min(min_ts, file.getModificationTime())
    return {'number_of_files': file_count, 'total_size': total_size, 'oldest_timestamp': min_ts}


def _is_running_on_dbr(spark: SparkSession) -> bool:
    flag = True
    try:
        # a canary class, randomly chosen
        spark._jvm.com.DatabricksMain.NONFATAL()
    except TypeError:
        flag = False
    return flag


def _snapshot_transactions(delta_table: DeltaTable):
    delta_log = _delta_log(delta_table)
    return delta_log.snapshot().transactions()


def _delta_log(delta_table: DeltaTable):
    spark = delta_table.toDF().sparkSession
    location = delta_table.detail().collect()[0]['location']
    is_databricks = _is_running_on_dbr(spark)
    if is_databricks:
        delta_log = spark._jvm.org.apache.spark.sql.delta.DeltaLog.forTable(
            spark._jsparkSession,
            location,
        )
    else:
        delta_log = spark._jvm.org.apache.spark.sql.delta.DeltaLog.forTable(
            spark._jsparkSession,
            location,
        )
    return delta_log


def _snapshot_allfiles(delta_table: DeltaTable) -> DataFrame:
    delta_log = _delta_log(delta_table)
    return DataFrame(delta_log.snapshot().allFiles(), delta_table.toDF().sparkSession)


class _DetailOutput:
    def __init__(self, delta_table: DeltaTable):
        detail_output = delta_table.detail().collect()[0].asDict()
        self.created_at = detail_output['createdAt']
        self.description: str = detail_output['description']
        self.format = detail_output['format']
        self.id = detail_output['id']
        self.last_modified = detail_output['lastModified']
        self.location = detail_output['location']
        self.min_reader_version = detail_output['minReaderVersion']
        self.min_writer_version = detail_output['minWriterVersion']
        self.name: str = detail_output['name']
        self.num_files = detail_output['numFiles']
        self.partition_columns = detail_output['partitionColumns']
        self.properties = detail_output['properties']
        self.size = detail_output['sizeInBytes']

    def humanize(self):
        self.num_files = _humanize_number(self.num_files)
        self.size = _humanize_bytes(self.size)

    def to_dict(self):
        return copy.deepcopy(self.__dict__)


def _deduplicate(
        delta_table: DeltaTable,
        backup_path: str,
        keys: list[str] | str,
        tiebreaking_columns: list[str] = None,
) -> DeltaTable:
    """
    Removes duplicates from a Delta Lake table.

    USE THIS AT YOUR OWN RISK. ANY DATA LOSS PROBLEMS ARE ON YOU, DO NOT OPEN AN ISSUE. DO NOT SEND ME AN EMAIL. DO. NOT.

    The reason that this function is hidden is because of the inherent complexities of doing this kind of thing, especially at scale.

    During this deduplication process, you must turn off any other process that write to the table. This is due to possible concurrency issues, and, in the event of failure and rollback, possible data loss.

    :param delta_table:
    :param backup_path: A temporary location used to stage de-duplicated data. The location that `temp_path` points to needs to be empty.
    :param keys: A list of column names used to distinguish rows. The order of this list does not matter.
    :param tiebreaking_columns: A list of column names used for ordering. The order of this list matters, with earlier elements "weighing" more than lesser ones. The columns will be evaluated in descending order. In the event of a tie, you will get non-deterministic results.
    :return: The same Delta Lake table as **delta_table**.
    """
    if isinstance(keys, str):
        keys = [keys]
    detail_object = _DetailOutput(delta_table)
    target_location = detail_object.location
    table_version = delta_table.history().select('version').limit(1).collect()[0].asDict()['version']
    print(
        f'IF THIS OPERATION FAILS, RUN RESTORE TO {table_version} WITH deltaTable.restoreToVersion({table_version})',
    )
    df = delta_table.toDF()
    spark = df.sparkSession

    # Deduplicate the dataframe using `hydro.spark.deduplicate_dataframe`
    deduped = hs.deduplicate_dataframe(df, keys=keys, tiebreaking_columns=tiebreaking_columns)
    deduped.write.format('delta').save(backup_path)

    # To delete matching rows, we use MERGE. MERGE requires a condition that matches rows between source and target
    merge_key_condition = ' AND '.join(
        [f'source.{key} = target.{key}' for key in keys],
    )

    # Perform deletion of existing duplicates using MERGE
    delta_table.alias('target').merge(
        deduped.select(keys).distinct().alias('source'),  # do we need distinct here?
        merge_key_condition,
    ).whenMatchedDelete().execute()

    # Insert the de-duplicated data into the target Delta table
    spark.read.format('delta').load(backup_path).write.format('delta').mode(
        'append',
    ).save(target_location)
    return delta_table



from __future__ import annotations

from typing import Any
from uuid import uuid4

from delta import DeltaTable
from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from pyspark.sql import Window

import hydro.spark as hs
from hydro import _humanize_bytes
from hydro import _humanize_number
from hydro import _humanize_timestamp
from hydro._delta import _DetailOutput
from hydro._delta import _snapshot_allfiles
from hydro._delta import _snapshot_transactions
from hydro._delta import _summarize_data_files


def scd(
        delta_table: DeltaTable,
        source: DataFrame,
        keys: list[str] | str,
        effective_ts: str,
        end_ts: str = None,
        scd_type: int = 2,
) -> DeltaTable:
    """

    Executes a slowly changing dimensions transformation and merge.

    Supports Type 1 and Type 2 SCD.

    :param delta_table: The target Delta Lake table that is to be updated. See `hydro.delta.bootstrap_scd2` if you need to create an SCD2 table.
    :param source: The source data that will be used to merge with `delta_table`
    :param keys: Column name(s) that identify unique rows. Can be a single column name as a string, or a list of strings, where order of the list does not matter.
    :param effective_ts: The name of the existing column that will be used as the “start” or “effective” indicator for a given entity. The column be of any orderable type, including timestamp, date, string, and numeric types.
    :param end_ts: Only required for type 2. The name of the non-existing column that will be used as the “end” indicator for a given entity. Its type will match the type of effective_ts.
    :param scd_type: The type of SCD that is to be performed
    :return: The same `delta_table`

    Example
    -----

     For **`SCD Type 2 <https://en.wikipedia.org/wiki/Slowly_changing_dimension#Type_2:_add_new_row>`**,

     Given a Delta Lake table:

    .. code-block:: python

        +---+-----------+----------+----------+
        | id|   location|      date|  end_date|
        +---+-----------+----------+----------+
        |  1|      Kochi|2018-01-01|2019-01-01|
        |  1|Lake Forest|2019-01-01|      null|
        +---+-----------+----------+----------+

    And a source DataFrame:

    .. code-block:: python

        +---+--------+----------+
        | id|location|      date|
        +---+--------+----------+
        |  1|  Irvine|2020-01-01|
        |  2|  Venice|2022-01-01|
        +---+--------+----------+



    .. code-block:: python

        import hydro.delta as hd
        hd.scd(delta_table, df, ["id"], effective_ts="date", end_ts="end_date")

    Results in:

    .. code-block:: python

        +---+-----------+----------+----------+
        | id|   location|      date|  end_date|
        +---+-----------+----------+----------+
        |  1|      Kochi|2018-01-01|2019-01-01|
        |  1|Lake Forest|2019-01-01|2020-01-01|
        |  1|     Irvine|2020-01-01|      null|
        |  2|     Venice|2022-01-01|      null|
        +---+-----------+----------+----------+

    See Also
    -----

    bootstrap_scd2


    """
    if isinstance(keys, str):
        keys = [keys]

    def _scd2(
            delta_table: DeltaTable,
            source: DataFrame,
            keys: list[str] | str,
            effective_ts: str,
            end_ts: str,
    ):
        if not end_ts:
            raise ValueError(
                '`end_ts` parameter not provided, type 2 scd requires this',
            )

        updated_rows = delta_table.toDF().join(source, keys, 'left_semi').filter(F.col(end_ts).isNull())
        combined_rows = updated_rows.unionByName(source, allowMissingColumns=True)
        window = Window.partitionBy(keys).orderBy(effective_ts)
        final_payload = combined_rows.withColumn(
            end_ts,
            F.lead(effective_ts).over(window),
        )
        merge_keys = keys + [effective_ts]
        merge_key_condition = ' AND '.join(
            [f'source.{key} = target.{key}' for key in merge_keys],
        )
        delta_table.alias('target').merge(
            final_payload.alias('source'),
            merge_key_condition,
        ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()
        return delta_table

    def _scd1(
            delta_table: DeltaTable,
            source: DataFrame,
            keys: list[str] | str,
            effective_ts: str,
    ):
        window = Window.partitionBy(keys).orderBy(F.col(effective_ts).desc())
        row_number_uuid = uuid4().hex  # avoid column collisions by using uuid
        final_payload = (
            source.withColumn(
                row_number_uuid,
                F.row_number().over(window),
            )
            .filter(F.col(row_number_uuid) == 1)
            .drop(row_number_uuid)
        )
        merge_key_condition = ' AND '.join(
            [f'source.{key} = target.{key}' for key in keys],
        )
        delta_table.alias('target').merge(
            final_payload.alias('source'),
            merge_key_condition,
        ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()
        return delta_table

    if scd_type == 2:
        return _scd2(delta_table, source, keys, effective_ts, end_ts)
    elif scd_type == 1:
        return _scd1(delta_table, source, keys, effective_ts)
    else:
        raise ValueError('`scd_type` not of (1,2)')


def bootstrap_scd2(
        source_df: DataFrame,
        keys: list[str] | str,
        effective_ts: str,
        end_ts: str,
        table_properties: dict[str, str] = None,
        partition_columns: list[str] = None,
        comment: str = None,
        path: str = None,
        table_identifier: str = None,
) -> DeltaTable:
    """

    Creates an SCD2-ready Delta Lake table from a source DataFrame.

    :param source_df: Source data that will populate the final Delta Lake table
    :param keys: Column name(s) that identify unique rows. Can be a single column name as a string, or a list of strings, where order of the list does not matter.
    :param effective_ts: The name of the existing column that will be used as the "start" or "effective" timestamp for a given entity. The column be of any orderable type, including timestamp, date, string, and numeric types.
    :param end_ts: The name of the non-existing column that will be used as the "end" timestamp for a given entity. Its type will match the type of `effective_ts`.
    :param table_properties: A set of [Delta Lake table properties](https://docs.delta.io/latest/table-properties.html) or custom properties.
    :param partition_columns: A set of column names that will be used to partition the resulting Delta Lake table.
    :param comment: Comment that describes the table.
    :param path: Specify the path to the directory where table data is stored, which could be a path on distributed storage.
    :param table_identifier: The table name. Optionally qualified with a database name [catalog_name.] [database_name.] table_name.
    :return: The resulting DeltaTable object


    Example
    -----

    Given a DataFrame:

    .. code-block:: python

            +---+-----------+----------+
            | id|   location|      date|
            +---+-----------+----------+
            |  1|      Kochi|2018-01-01|
            |  1|Lake Forest|2019-01-01|
            +---+-----------+----------+

    Run bootstrap_scd2:

    .. code-block:: python

        import hydro.delta as hd
        delta_table = hd.bootstrap_scd2(
            df,
            keys=["id"],
            effective_ts="date",
            end_ts="end_date",
            path="/path/to/delta/table",
        )

    Results in the below data being persisted to the given path, in this case, `/path/to/delta/table`

    .. code-block:: python

            +---+-----------+----------+----------+
            | id|   location|      date|  end_date|
            +---+-----------+----------+----------+
            |  1|      Kochi|2018-01-01|2019-01-01|
            |  1|Lake Forest|2019-01-01|      null|
            +---+-----------+----------+----------+

    See Also
    -----

    scd

    """

    if partition_columns is None:
        partition_columns = []
    if not path and not table_identifier:
        raise ValueError(
            'Need to specify one (or both) of `path` and `table_identifier`',
        )

    # scd2-ify the existing data
    window = Window.partitionBy(keys).orderBy(effective_ts)
    final_payload = source_df.withColumn(
        end_ts,
        F.lead(effective_ts).over(window),
    )

    # build the DeltaTable object
    builder = DeltaTable.createOrReplace(
        source_df.sparkSession,
    )  # TODO change to createIfNotExists?
    if table_properties:
        for k, v in table_properties.items():
            builder = builder.property(k, v)
    builder = builder.addColumns(source_df.schema)
    builder = builder.partitionedBy(*partition_columns)
    if comment:
        builder = builder.comment(comment)
    if path:
        builder = builder.location(path)
    if table_identifier:
        builder = builder.tableName(table_identifier)
    builder.execute()  # save DeltaTable to disk

    # write data to disk
    if table_identifier:
        final_payload.write.format('delta').option('mergeSchema', 'true').mode(
            'append',
        ).saveAsTable(table_identifier)
        delta_table = DeltaTable.forName(source_df.sparkSession, table_identifier)
    else:
        final_payload.write.format('delta').option('mergeSchema', 'true').mode(
            'append',
        ).save(path)
        delta_table = DeltaTable.forPath(source_df.sparkSession, path)
    return delta_table


def partial_update_set(
        delta_frame: DataFrame | DeltaTable,
        source_alias: str = 'source',
        target_alias: str = 'target',
) -> F.col:
    """
    Generates an update set for a Delta Lake MERGE operation where the source data provides partial updates.

    Partial updates in this case are when some columns in the data are NULL, but are meant to be non-destructive, or is there no semantic meaning to the NULLs.

    :param delta_frame: A DeltaTable or DataFrame that describes a source MERGE dataframe
    :param source_alias: A temporary name given to the source data of the MERGE
    :param target_alias: A temporary name given to the target Delta Table of the MERGE
    :return: A dictionary that describes non-destructive updates for all fields in `delta_frame` in {key: coalesce(source.key, target.key)} form

    Example
    -----

    **Example**

    Given a Delta Lake table:

    .. code-block:: python

        +---+----------+------+
        | id|  location|status|
        +---+----------+------+
        |  1|california|  null|
        +---+----------+------+

    And some source data that will partially update the Delta Lake table:

    .. code-block:: python

        +---+--------+------+
        | id|location|status|
        +---+--------+------+
        |  1|    null|active|
        +---+--------+------+

    Perform the following `MERGE`:

    .. code-block:: python

        import hydro.delta as hd
        delta_table.alias("target").merge(
            df.alias("source"), "source.id = target.id"
        ).whenNotMatchedInsertAll().whenMatchedUpdate(set=hd.partial_update_set(df)).execute()

    With the resulting Delta Lake table:

    .. code-block:: python

        +---+----------+------+
        | id|  location|status|
        +---+----------+------+
        |  1|california|active|
        +---+----------+------+


    """
    # why does cov lie?
    if isinstance(delta_frame, DeltaTable):
        delta_frame = delta_frame.toDF()
    fields = hs.fields(delta_frame)
    return {field: F.coalesce(f'{source_alias}.{field}', f'{target_alias}.{field}') for field in fields}


def file_stats(delta_table: DeltaTable) -> DataFrame:
    """

    Returns detailed information about the files of the current snapshot of a Delta Lake table, including (per-file):

    - name of file
    - size of file
    - partition values
    - modification time
    - is data change
    - statistics (min, max, and null counts)
    - tags

    This is done by scanning the table's transaction log, so it is fast, cheap, and scalable.

    :return: A DataFrame that describes the physical files that compose a given Delta Lake table
    """
    return _snapshot_allfiles(delta_table)


def partition_stats(delta_table: DeltaTable) -> DataFrame:
    """

    Returns detailed information about the partitions of the current snapshot of a Delta Lake table including (per-partition):

    - total size in bytes
    - file byte size quantiles (0, .25, .5, .75, 1.0) - where 0 is min and 1.0 is max
    - total number of records
    - total number of files
    - oldest and newest timestamps

    This is done by scanning the table's transaction log, so it is fast, cheap, and scalable.

    :return: A DataFrame that describes the size of all partitions in the table
    """
    allfiles = _snapshot_allfiles(delta_table)
    detail = _DetailOutput(delta_table)
    # the snapshot encodes partition values in the top-level `partitionValues` column
    partition_columns = [f'partitionValues.{col}' for col in detail.partition_columns]
    return allfiles.groupBy(*partition_columns).agg(
        F.sum('size').alias('total_bytes'),
        F.percentile_approx('size', [0, 0.25, 0.5, 0.75, 1.0]).alias('bytes_quantiles'),
        F.sum(F.get_json_object('stats', '$.numRecords')).alias('num_records'),
        F.count('*').alias('num_files'),
        F.min('modificationTime').alias('oldest_timestamp'),
        F.max('modificationTime').alias('newest_timestamp'),
    )


def zordering_stats(delta_table: DeltaTable) -> DataFrame:
    """

    Returns a DataFrame that describes the Z-Ordering that has been applied to the table.

    The resulting DataFrame has schema of `zOrderBy`, `count`.

    :return: A DataFrame with schema `zOrderBy`, `count`.
    """
    return (
        delta_table.history()
        .filter("operation == 'OPTIMIZE'")
        .filter('operationParameters.zOrderBy IS NOT NULL')
        .select('operationParameters.zOrderBy')
        .groupBy('zOrderBy')
        .count()
    )


def detail(delta_table: DeltaTable) -> dict[str, Any]:
    """

    Returns details about a Delta Lake table including:

    - table created timestamp
    - description
    - table format
    - table id
    - table last modified
    - location
    - minimum reader version
    - minimum writer version
    - table name
    - total number of records of current snapshot
    - total number of files of current snapshot
    - partition columns
    - total number of partitions of the current snapshot
    - properties
    - total data size of the current snapshot
    - percentage collected stats of the current snapshot
    - snapshot version

    :return: A dictionary representing enhanced details of a Delta Lake table
    """
    detail_output = _DetailOutput(delta_table)
    detail_output.humanize()
    details = detail_output.to_dict()
    allfiles = _snapshot_allfiles(delta_table)
    partition_columns = [f'partitionValues.{col}' for col in details['partition_columns']]

    # compute the number of records in the current snapshot by scanning the Delta log
    num_records = (
        allfiles.select(
            F.get_json_object('stats', '$.numRecords').alias('num_records'),
        )
        .agg(F.sum('num_records').alias('num_records'))
        .collect()[0]['num_records']
    )
    details['numRecords'] = _humanize_number(num_records)

    # compute the percentage of files that have statistics in the current snapshot by scanning the Delta log
    stats_percentage = (
        allfiles.agg(
            F.avg(
                F.when(F.col('stats').isNotNull(), F.lit(1)).otherwise(F.lit(0)),
            ).alias('stats_percentage'),
        )
    ).collect()[0]['stats_percentage']
    details['stats_percentage'] = stats_percentage * 100

    # compute the number of partitions in the current snapshot by scanning the Delta log
    partition_count = allfiles.select(*partition_columns).distinct().count()
    details['partition_count'] = _humanize_number(partition_count)

    # compute the version of the current snapshot by looking at the table's history, which scans the Delta log
    version = delta_table.history().select('version').limit(1).collect()[0].asDict()['version']
    details['version'] = _humanize_number(version)
    return details


def summarize_all_files(delta_table: DeltaTable, humanize: bool = True) -> dict[str, str]:
    """

    Lists and summarizes all of the contents of a Delta Lake table's data directory.

    The directory will contain:

    - data files that are part of the current snapshot
    - data files that are "tombstoned" and not part of the current snapshot

    Returns summary statistics including:

    - the total number of files
    - the total size of the files
    - the oldest timestamp of the files

    :param: humanize: Whether or not the results should be made more easily read by humans. Turn this to False if you're looking to do calculations on the raw metrics.
    :returns: A dictionary containing summary statistics about all of the data files under the Delta Lake table's location
    """
    summary = _summarize_data_files(delta_table)
    if humanize:
        summary['number_of_files'] = _humanize_number(summary['number_of_files'])
        summary['total_size'] = _humanize_bytes(summary['total_size'])
        summary['oldest_timestamp'] = _humanize_timestamp(summary['oldest_timestamp'])
    return summary


def idempotency_markers(delta_table: DeltaTable) -> str:
    """

    Exposes a Delta Lake table's idempotency markers, i.e txnAppId and txnVersion.

    Currently this returns a string representation of a Scala map.

    :returns: A string that represents a Scala map of the idempotency markers. This map is in the form Map(key-> value) where key is a given txnAppId and value is the associated version.
    """
    return _snapshot_transactions(delta_table)



from __future__ import annotations

import hashlib
from collections import Counter
from typing import Callable
from uuid import uuid4

import pyspark.sql.functions as F
from pyspark.sql import Column
from pyspark.sql import DataFrame
from pyspark.sql import Window
from pyspark.sql.types import DataType
from pyspark.sql.types import StructType

from hydro._spark import _create_drop_field_column
from hydro._spark import _field_trie
from hydro._spark import _fields
from hydro._spark import _get_fields_by_regex
from hydro._spark import _get_fields_by_type
from hydro._spark import _map_fields


def fields(df: DataFrame) -> list[str]:
    """

    Returns a list of names of all fields of a DataFrame, including nested ones.

    :param df: DataFrame that you want to extract all fields from
    :return: A list of column names, all strings

    Example
    -----

    .. code-block:: python

        import hydro.spark as hs
        df = spark.range(1)
        hs.fields(df)

    .. code-block:: python

        ['id']
    """
    return _fields(df, False)


def fields_with_types(df: DataFrame) -> list[tuple[str, DataType]]:
    """

    Returns a list of tuples of names and types of all fields of a DataFrame, including nested ones.

    :param df: DataFrame that you want to extract all fields and types from
    :return: A list of tuples of (column_name, type)

    Example
    -----

    .. code-block:: python

        import hydro.spark as hs
        df = spark.range(1)
        hs.fields_with_types(df)

    .. code-block:: python

        [('id', LongType())]


    """
    return _fields(df, True)


def deduplicate_dataframe(
    df: DataFrame,
    keys: list[str] | str = None,
    tiebreaking_columns: list[str] = None,
) -> DataFrame:
    """
    Removes duplicates from a Spark DataFrame.

    :param df: The target Delta Lake table that contains duplicates.
    :param keys: A list of column names used to distinguish rows. The order of this list does not matter. If not provided, will operate the same as `Dataframe.drop_duplicates()`
    :param tiebreaking_columns: A list of column names used for ordering. The order of this list matters, with earlier elements "weighing" more than lesser ones. The columns will be evaluated in descending order. If not provided, will will operate the same as `Dataframe.drop_duplicates(keys)`
    :return: The deduplicated DataFrame

    Example
    -----

    Given an input DataFrame

    .. code-block:: python

        +---+-----+-----------------------+
        |id |type |ts                     |
        +---+-----+-----------------------+
        |1  |watch|2023-01-09 15:48:00.000|
        |1  |watch|2023-01-09 15:48:00.001|
        +---+-----+-----------------------+

    There are two events with the same primary key `id`, but with a slightly different timestamp. Two rows should not share a primary key. A common way of dealing with this is to break the "tie" based on some orderable column(s).

    This can be done with the following code:


    .. code-block:: python

        import hydro.spark as hs
        data  = [{"id": 1, "type": "watch", "ts": "2023-01-09 15:48:00.001"}, {"id": 1, "type": "watch", "ts": "2023-01-09 15:48:00.001"}]
        df = spark.createDataFrame(data)
        hs.deduplicate_dataframe(df, ["id"], ["ts"])

    Results in

    .. code-block:: python

        +---+-----------------------+-----+
        |id |ts                     |type |
        +---+-----------------------+-----+
        |1  |2023-01-09 15:48:00.001|watch|
        +---+-----------------------+-----+

    """
    if keys is None:
        keys = []

    if tiebreaking_columns is None:
        tiebreaking_columns = []

    if isinstance(keys, str):
        keys = [keys]

    if not keys:
        return df.drop_duplicates()

    if df.isStreaming and tiebreaking_columns:
        print('df is streaming, ignoring `tiebreaking_columns`')

    count_col = uuid4().hex  # generate a random column name that is virtually certain to not be in the dataset
    window = Window.partitionBy(keys)

    # noinspection PyTypeChecker
    dupes = df.withColumn(count_col, F.count('*').over(window)).filter(F.col(count_col) > 1).drop(count_col)
    if tiebreaking_columns and not df.isStreaming:
        row_number_col = uuid4().hex
        tiebreaking_desc = [F.col(col).desc() for col in tiebreaking_columns]  # potential enhancement here
        tiebreaking_window = window.orderBy(tiebreaking_desc)
        deduped = (
            dupes.withColumn(row_number_col, F.row_number().over(tiebreaking_window))  # row_number is non-deterministic in the event of ties
            .filter(F.col(row_number_col) == 1)  # take the top row
            .drop(row_number_col)
        )
    else:
        deduped = dupes.drop_duplicates(keys)
    return deduped


def hash_fields(df: DataFrame, denylist_fields: list[str] = None, algorithm: str = 'xxhash64', num_bits=256) -> Column:
    """

    Generates a hash digest of all fields.

    :param df: Input dataframe that is to be hashed.
    :param denylist_fields: Fields that will not be hashed.
    :param algorithm: The function that is used to generate the hash digest, includes:

            * ``xxhash64`` (default) :class:`pyspark.sql.functions.xxhash64`
            * ``md5`` :class:`pyspark.sql.functions.md5`
            * ``sha1`` :class:`pyspark.sql.functions.sha1`
            * ``sha2`` :class:`pyspark.sql.functions.sha2`
            * ``hash`` :class:`pyspark.sql.functions.hash`
    :param num_bits: Only for sha2. The desired bit length of the result.
    :return: A column that represents the hash.

    Example
    -----

    Given an input DataFrame

    .. code-block:: python

        +---+---+-----+
        | id| ts| type|
        +---+---+-----+
        |  1|  1|watch|
        |  1|  1|watch|
        |  1|  2|watch|
        +---+---+-----+

    Row hashes are very helpful and convenient when trying to compare rows, especially rows with a lot of columns.

    .. code-block:: python

        import hydro.spark as hs
        data  = [{"id": 1, "type": "watch", "ts": "1"}, {"id": 1, "type": "watch", "ts": "1"}, {"id": 1, "type": "watch", "ts": "2"}]
        df = spark.createDataFrame(data)
        df.withColumn("row_hash", hs.hash_fields(df))

    Results in

    .. code-block:: python

        +---+-----+---+-------------------+
        | id| type| ts|           row_hash|
        +---+-----+---+-------------------+
        |  1|watch|  1|8553228534919528539|
        |  1|watch|  1|8553228534919528539|
        |  1|watch|  2|8916583907181700702|
        +---+-----+---+-------------------+

    The rows with identical content have identical hash values. Rows with different content have different hash values.

    This is very helpful when comparing rows. Instead of typing col1 = col1 AND col2 = col2 ..., hash = hash can be used. This saves developer time and keystrokes.

    """
    supported_algorithms = ['sha1', 'sha2', 'md5', 'hash', 'xxhash64']

    if algorithm not in supported_algorithms:
        raise ValueError(f'Algorithm {algorithm} not in supported algorithms {supported_algorithms}')

    all_fields = fields(df)

    if denylist_fields:
        all_fields = list(set(all_fields) - set(denylist_fields))

    all_fields.sort()
    if algorithm == 'sha1':
        hash_col = F.sha1(F.concat_ws('', *all_fields))
    elif algorithm == 'sha2':
        hash_col = F.sha2(F.concat_ws('', *all_fields), num_bits)
    elif algorithm == 'hash':
        hash_col = F.hash(F.concat_ws('', *all_fields))
    elif algorithm == 'xxhash64':
        hash_col = F.xxhash64(F.concat_ws('', *all_fields))
    else:
        hash_col = F.md5(F.concat_ws('', *all_fields))

    return hash_col


def hash_schema(df: DataFrame, denylist_fields: list[str] = None) -> Column:
    """

    Generates a hash digest of a DataFrame's schema. Uses the hashlib.md5 algorithm.

    :param df: Input dataframe whose schema is to be hashed.
    :param denylist_fields: Fields that will not be hashed.
    :return: A column that represents the hash.

    Example
    -----

    .. code-block:: python

        import hydro.spark as hs
        df = spark.range(1)
        df.withColumn("schema_hash", hs.hash_schema(df))

    .. code-block:: python

        +---+--------------------------------+
        |id |schema_hash                     |
        +---+--------------------------------+
        |0  |b80bb7740288fda1f201890375a60c8f|
        +---+--------------------------------+

    The schema of tables can change, and sometimes it is helpful to be able to determine if a table's schema has changed.


    """

    all_fields = fields(df)
    if denylist_fields:
        all_fields = list(set(all_fields) - set(denylist_fields))

    fields_set = set(all_fields)
    if len(all_fields) != len(fields_set):
        dupes = [item for item, count in Counter(all_fields).items() if count > 1]
        raise ValueError(f'Duplicate field(s) detected in df, {dupes}')

    """
    ChatGPT 🤖 prompt:
     python's hash function seems to not be deterministic across sessions. give me a python program that gives the md5 hash of a string (python 3)
    """
    schema_hash = hashlib.md5(''.join(sorted(all_fields)).encode('utf-8')).hexdigest()
    hash_col = F.lit(schema_hash)  # amalgamate list as string bc list is un-hashable
    return hash_col


def map_fields(df: DataFrame, field_list: list[str], function: Callable) -> DataFrame:
    """

    Applies a function `function` over fields that are specified in a list.

    :param df:
    :param field_list: A list of fields the function is to be applied to.
    :param function: Any `pyspark.sql.function` or lambda function that takes a column.
    :return:

    Example
    -----

    .. code-block:: python

        data = [{"empty": "   ", "ibm": "   🥴  ", "other": "thing"}]
        df = spark.createDataFrame(data)
        df.show()

    .. code-block:: python

        +-----+-------+-----+
        |empty|    ibm|other|
        +-----+-------+-----+
        |     |   🥴  |thing|
        +-----+-------+-----+

    .. code-block:: python

        hs.map_fields(df, ["empty", "ibm"], F.trim)

    .. code-block:: python

        +-----+---+-----+
        |empty|ibm|other|
        +-----+---+-----+
        |     | 🥴|thing|
        +-----+---+-----+

    A lambda function can be used to compose functions:

    .. code-block:: python


        hs.map_fields(df, ["empty", "ibm"], lambda x: F.when(F.trim(x) == F.lit(""), None).otherwise(F.trim(x)))

    .. code-block:: python

        +-----+---+-----+
        |empty|ibm|other|
        +-----+---+-----+
        | null| 🥴|thing|
        +-----+---+-----+

    F.expr() can be used via interpolation with f-strings (it's a little hard to read and write, though):

    .. code-block:: python

        hs.map_fields(df, ["empty", "ibm"], lambda x: F.expr(f"nullif(trim({x}), '')"))

    .. code-block:: python

        +-----+---+-----+
        |empty|ibm|other|
        +-----+---+-----+
        | null| 🥴|thing|
        +-----+---+-----+

    """
    return _map_fields(df, field_list, function)


def map_fields_by_regex(df: DataFrame, regex: str, function: Callable) -> DataFrame:
    """

    Applies a function `function` over fields that match a regular expression.


    :param df:
    :param regex: Regular expression pattern. Uses Python's `re` module.
    :param function: Any `pyspark.sql.function` or lambda function that takes a column.
    :return: Resulting DataFrame


    See Also
    -----

    map_fields

    """

    matches = _get_fields_by_regex(df, regex)
    return _map_fields(df, matches, function)


def map_fields_by_type(df: DataFrame, target_type: DataType, function: Callable) -> DataFrame:
    """

    Applies a function `function` over fields that have a target type.

    :param df:
    :param target_type:
    :param function: Any `pyspark.sql.function` or lambda function that takes a column.
    :return: Resulting DataFrame

    See Also
    -----

    map_fields

    """
    pertinent_fields = _get_fields_by_type(df, target_type)
    return _map_fields(df, pertinent_fields, function)


def select_fields_by_type(df: DataFrame, target_type: DataType):
    """

    Selects fields according to a provided regular expression pattern. Works with nested fields (but un-nests them)


    :param df:
    :param target_type: A DataType type that is to be selected.
    :return: A new DataFrame with the selected fields

    Example
    -----

    .. code-block:: python

        data = [{"string1": "one", "int": 1, "string2": "two"}]
        df = spark.createDataFrame(data)

    .. code-block:: python

        import hydro.spark as hs
        import pyspark.sql.types as T
        hs.select_fields_by_type(df, T.StringType())

    .. code-block:: python

        +-------+-------+
        |string1|string2|
        +-------+-------+
        |    one|    two|
        +-------+-------+

    """
    pertinent_fields = _get_fields_by_type(df, target_type)
    return df.select(*pertinent_fields)


def select_fields_by_regex(df: DataFrame, regex: str) -> DataFrame:
    """

    Selects fields according to a provided regular expression pattern. Works with nested fields (but un-nests them)

    :param df:
    :param regex: Regular expression pattern. Uses Python's `re` module.
    :return: A new DataFrame with the selected fields


    .. code-block:: python

        data = [{"string1": "one", "int": 1, "string2": "two"}]
        df = spark.createDataFrame(data)

    .. code-block:: python

        import hydro.spark as hs
        import pyspark.sql.types as T
        hs.select_fields_by_regex(df, "string.*")

    .. code-block:: python

        +-------+-------+
        |string1|string2|
        +-------+-------+
        |    one|    two|
        +-------+-------+

    """
    matches = _get_fields_by_regex(df, regex)
    return df.select(*matches)


def drop_fields(df: DataFrame, fields_to_drop: list[str]) -> DataFrame:
    """

    Drops a DataFrame's fields, including nested fields and top-level columns.

    :param df:
    :param fields_to_drop: A list of field names that are to be dropped
    :return: A new DataFrame without the specified fields

    Example
    -----

    .. code-block:: python

        # This is a silly way of creating a nested DataFrame, it's here for brevity
        row = Row(nest=Row(key="val", society="spectacle"))
        df = spark.createDataFrame([row])
        df.printSchema()

    And here is the schema:

    .. code-block:: python

        root
         |-- nest: struct (nullable = true)
         |    |-- key: string (nullable = true)
         |    |-- society: string (nullable = true)

    Using hydro, we can simply drop nested fields.

    .. code-block:: python

        import hydro.spark as hs
        hs.drop_fields(df, ["nest.key"])

    With the resulting schema:

    .. code-block:: python

        root
         |-- nest: struct (nullable = true)
         |    |-- society: string (nullable = true)

    """
    if isinstance(fields_to_drop, str):
        fields_to_drop = [fields_to_drop]
    tries = _field_trie(fields_to_drop)
    for trie in tries.items():
        if trie[1] == [None]:
            df = df.drop(trie[0])
        else:
            name, col = _create_drop_field_column(trie)
            df = df.withColumn(name, col)
    return df


def infer_json_field(df: DataFrame, target_field: str, options: dict[str, str] = None) -> StructType:
    """

    Parses a JSON string and infers its schema.

    :param df:
    :param target_field: A field that contains JSON strings that are to be inferred.
    :param options: Standard JSON reader options, including `header`. See :class:pyspark.sql.DataFrameReader.json
    :return: The inferred StructType

    Example
    -----

    .. code-block:: python

         data = [{"id": 1, "payload": '{"salt": "white"}'}, {"id": 2, "payload": '{"pepper": "black"}'}]
         df = spark.createDataFrame(data)
         df.show()

    Looks like:

    .. code-block:: python

        +---+-------------------+
        | id|            payload|
        +---+-------------------+
        |  1|  {"salt": "white"}|
        |  2|{"pepper": "black"}|
        +---+-------------------+

    But there's a problem, our schema doesn't include salt and pepper.

    .. code-block:: python

        root
         |-- id: long (nullable = true)
         |-- payload: string (nullable = true)

    We can use hydro to fix this:

    .. code-block:: python

        import hydro.spark as hs
        schema = hs.infer_json_field(df, "payload")

    And use the resulting schema to parse the fields:

    .. code-block:: python

        df.withColumn("payload", F.from_json("payload", schema))

    With the new schema:

    .. code-block:: python

        root
         |-- id: long (nullable = true)
         |-- payload: struct (nullable = true)
         |    |-- pepper: string (nullable = true)
         |    |-- salt: string (nullable = true)

    Notice how there are separate fields for salt and pepper. And now these are addressable leaf nodes.

    """
    if not options:
        options = dict()
    spark = df.sparkSession
    rdd = df.select(target_field).rdd.map(lambda row: row[0])
    return spark.read.options(**options).json(rdd).schema


def infer_csv_field(df: DataFrame, target_field: str, options: dict[str, str] = None) -> StructType:
    """

    Parses a CSV string and infers its schema.

    :param df:
    :param target_field: A field that contains CSV strings that are to be inferred.
    :param options: Standard CSV reader options, including `header`. See :class:pyspark.sql.DataFrameReader.csv
    :return: The inferred StructType

    Example
    -----

    .. code-block:: python

        data = {
            'id': 1, 'payload': '''type,date
            watch,2023-01-09
            watch,2023-01-10
        ''',
        }
        df = spark.createDataFrame([data])
        df.show()


    .. code-block:: python

        +---+-----------------------------------------------+
        |id |payload                                        |
        +---+-----------------------------------------------+
        |1  |type,date\\nwatch,2023-01-09\\nwatch,2023-01-10\\n|
        +---+-----------------------------------------------+


    .. code-block:: python

        import hydro.spark as hs
        csv_options = {'header': 'True'}
        schema = hs.infer_csv_field(df, 'payload', options=csv_options)
        schema.simpleString()

    .. code-block:: python

        'struct<type:string,date:string>'


    """
    if not options:
        options = dict()
    spark = df.sparkSession
    rdd = df.select(target_field).rdd.map(lambda row: row[0])
    # noinspection PyTypeChecker
    return spark.read.options(**options).csv(rdd).schema



from __future__ import annotations

import re
from collections import defaultdict
from copy import copy
from typing import Callable

from pyspark.sql import Column
from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from pyspark.sql.types import DataType
from pyspark.sql.types import StructType


class _DeconstructedField:
    def __init__(self, field: str):
        split_field = field.split('.')
        self.levels = copy(split_field)
        self.trunk = split_field.pop(0)
        if len(split_field) > 1:
            *self.branches, self.leaf = split_field
        elif len(split_field) == 0:
            self.branches, self.leaf = [], None
        else:
            self.branches, self.leaf = [], split_field[0]
        self.trunk_and_branches = '.'.join([self.trunk] + self.branches)


def _field_trie(fields: list[str]):
    result = defaultdict(list)
    for field in fields:
        deconstructed_field = _DeconstructedField(field)
        trunk_and_branches = deconstructed_field.trunk_and_branches
        leaf = deconstructed_field.leaf
        result[trunk_and_branches].append(leaf)
    return result


def _fields(
    df: DataFrame,
    include_types: bool,
) -> list[tuple[str, DataType] | str]:
    # ChatGPT 🤖 prompt:
    # write a program that takes a PySpark StructType and returns the leaf node field names, even the nested ones # noqa: E501
    schema = df.schema

    def get_leaf_fields(
        struct: StructType,
        include_types: bool,
    ) -> list[tuple[str, DataType] | str]:
        def _get_leaf_fields(
            struct: StructType,
            prefix: str,
        ) -> list[tuple[str, DataType] | str]:
            fields: list[tuple[str, DataType] | str] = []
            for field in struct:
                if isinstance(field.dataType, StructType):
                    fields.extend(
                        _get_leaf_fields(
                            field.dataType,
                            prefix + field.name + '.',
                        ),
                    )
                else:
                    if include_types:
                        fields.append((prefix + field.name, field.dataType))
                    else:
                        fields.append(prefix + field.name)
            return fields

        return _get_leaf_fields(struct, '')

    return get_leaf_fields(schema, include_types)


def _get_fields_by_regex(df: DataFrame, regex: str) -> list[str]:
    # ChatGPT 🤖 prompt:
    # i have a regex pattern string. write a python program that iterates through a list of strings and returns elements that match the regex
    regex = re.compile(regex)
    all_fields = _fields(df, False)
    matches = [field for field in all_fields if regex.search(field)]
    return matches


def _get_fields_by_type(df: DataFrame, target_type: DataType) -> list[str]:
    all_fields = _fields(df, True)
    pertinent_fields = [field[0] for field in all_fields if field[1] == target_type]
    return pertinent_fields


def _map_fields(df: DataFrame, fields_to_map: list[str], function: Callable) -> DataFrame:
    for field in fields_to_map:
        df = df.withColumn(field, function(field))
    return df


def _create_drop_field_column(fields_to_drop: tuple[str, list[str | None]]) -> tuple[str, Column]:

    address, leaves = fields_to_drop

    if not leaves or leaves[0] is None:
        raise ValueError(f'Cannot drop top-level field `{address}` with this function. Use df.drop() instead.')

    def _traverse_nest(nest, l, c=0):
        if len(l) == 0:  # termination condition
            return F.col(nest).dropFields(*leaves)
        else:  # recursive step
            current_level = l[0]
            return F.col(nest).withField(current_level, _traverse_nest(f'{nest}.{current_level}', l[1:], c + 1))

    levels = address.split('.')
    if len(levels) == 1:
        return address, F.col(address).dropFields(*leaves)
    col = _traverse_nest(levels[0], levels[1:])
    return levels[0], col
